#! /usr/local/bin/perl
# ESA (C) 2000-2018 
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#
# Aimed to help building DAL_C_lib shared library.
#
# $Id: Makefile.PL,v 1.13 2013/04/01 11:42:20 eojero Exp $
#

use Env;
use ExtUtils::MakeMaker;
use FileHandle;
use File::Basename;
use Config; # for $Config{ccflags}

# call WriteMakefile from ExtUtils::MakeMaker

WriteMakefile(
              'NAME'	  => 'DAL_C_lib',
              'OBJECT'    => 'DAL_C_lib_wrap.o',
	      'CONFIGURE' => \&initialize,
              'LD'        => '$(CC)',
              'CCFLAGS'   => '-DPERL_POLLUTE '.$Config{ccflags}
);

# initialize function

sub initialize {

    my $self = {};
    local $\ = "\n";

    die "Have you defined SAS_DIR?\n" if not defined $SAS_DIR;

    # Depending on the OS, select the proper flags to create the shared library
    # specially for Darwin.
    # Flags not defined here will be taken from perl configuration.

    if ($^O eq 'darwin') {
      $self->{'LDDLFLAGS'} = '-dynamiclib -single_module -flat_namespace -undefined suppress -Wl,-install_name -Wl,$(LIBDIR)/$@';
      $self->{'LDFLAGS'}   = '-fstack-protector';
      $self->{'DLEXT'}     = 'dylib';
    }
    if ($^O eq 'linux') {
      $self->{'LDDLFLAGS'} = '-shared -Wl,-install_name -Wl,$(LIBDIR)/$@';
      $self->{'DLEXT'}     = 'so';
    }



    my $fh = new FileHandle;
    $fh->open("../../VERSION", "r") or die $!;
    my $ver = <$fh>;
    chomp $ver;
    undef $fh;
    $self->{'VERSION'} = $ver;

    $self->{'INC'} = "-I$SAS_DIR/include -I$SAS_DIR/include/dalc -I$SAS_DIR/include/dal -I$SAS_DIR/include/utils";

    # If DECEIT env variable is set, then use only libdalc library,
    # otherwise use libdal, libutils and liberror.

    if($DECEIT){
	print "\nDAL_C_lib will be built for use with deceit\n";
	$self->{'LIBS'} = ["-L$SAS_DIR/lib -ldalc"];
    } else {
	print "\nDAL_C_lib will be built for stand alone use\n";
	
	my $libs = "-L$SAS_DIR/lib -ldal -lutils -lerror";
        # cfitsio lib is required
        my $libcfitsio = find_library('cfitsio');
        $libs = $libs . " " . $libcfitsio if ( $libcfitsio );
        my $libstdcpp = find_egcs_library('stdc++') ; 
	$libs = $libs . " " . $libstdcpp if ( $libstdcpp );
	$self->{'LIBS'} = [$libs];
    }
	
    return $self;
}

# -- these to be moved somewhere else

# find a header file in C_INCLUDE_PATH or CPLUS_INCLUDE_PATH
# return -I$dir 
sub find_header {
    my $header = shift;
    my @c_include_path = split /:/, $C_INCLUDE_PATH;
    my @cplus_include_path = split /:/, $CPLUS_INCLUDE_PATH;
    foreach my $dir (@c_include_path)
    {
	if(-e "$dir/$header")
	{
	    return "-I$dir";
	}
    }
    foreach my $dir (@cplus_include_path)
    {
	if(-e "$dir/$header")
	{
	    return "-I$dir";
	}
    }
    print "Error: Required $header not found. Build of this module may fail.\n";
}



# find a library in (DY)LD_LIBRARY_PATH or LIBRARY_PATH
# return -L$dir -l$lib
sub find_library {    
    my $lib = shift;
    if ($^O eq 'darwin') {
      my @ld_library_path = split /:/, $DYLD_LIBRARY_PATH;
    } else { 
      my @ld_library_path = split /:/, $LD_LIBRARY_PATH;
    }
    my @library_path = split /:/, $LIBRARY_PATH;
    foreach my $dir (@ld_library_path)
    {
	if(-e "$dir/lib$lib.so" or -e "$dir/lib$lib.dylib" or -e "$dir/lib$lib.a")
	{
            print "$lib found in $dir\n";
	    return "-L$dir -l$lib";
	}
    }
    foreach my $dir (@library_path)
    {
	if(-e "$dir/lib$lib.so" or -e "$dir/lib$lib.dylib" or -e "$dir/lib$lib.a")
	{
            print "$lib found in $dir\n";
	    return "-L$dir -l$lib";
	}
    }
    print "Error: Required $lib not found. Build of this module may fail.\n";
}

# find g++ library
# return -L$dir -l$lib
sub find_egcs_library {
    my $lib = shift;
    my $liba = `g++ -print-file-name=lib$lib.a`;
    my $libso = `g++ -print-file-name=lib$lib.so`;
    my $libdylib = `g++ -print-file-name=lib$libdylib.dylib`;
    chomp($liba);
    chomp($libso);
    chomp($libdylib);
    if(-e $liba)
    {
	my $l = quotemeta($lib);
	$liba =~ m{(/.+/)lib$l\.a};
        print "$lib found in $1\n";
	return "-L$1 -l$lib";
    } elsif(-e $libso)
    {
	my $l = quotemeta($lib);
	$libso =~ m{(/.+/)lib$l\.so};
        print "$lib found in $1\n";
	return "-L$1 -l$lib";
    } elsif(-e $libdylib)
    {
        my $l = quotemeta($lib);
        $libdylib =~ m{(/.+/)lib$l\.dylib};
        print "$lib found in $1\n";
        return "-L$1 -l$lib";
    } else {
	print "Error: Required $lib not found. Build of this module may fail.\n";
    }
}
